unit Main;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, system.Math, Vcl.StdCtrls, Vcl.Buttons,
  Vcl.ExtCtrls, VclTee.TeeGDIPlus, VCLTee.Series, VCLTee.TeEngine,
  VCLTee.TeeProcs, VCLTee.Chart;

type
  TForm1 = class(TForm)
    BitBtn1: TBitBtn;
    LabeledEdit1: TLabeledEdit;
    Memo1: TMemo;
    Chart1: TChart;
    Series1: TLineSeries;
    Series2: TPointSeries;
    procedure FormCreate(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
  private
    { Private 錾 }
    function Dirichlet_Eta_Function(s: double): double;
    function xDirichlet_Eta_Function(s: extended): extended;
    function Dirichlet_Eta_Star_Function(s: double): double;
    function xDirichlet_Eta_Star_Function(s : extended): extended;
    function Reflection_Coefficient(s: extended): extended;
    function Sum_Reverse_Order(s: extended): extended;
    function Alternating_Series_Convergence_Acceleration(s: extended): extended;
    function Duplication_Formula(two_x: extended): extended;
    function xGamma(x: extended): extended;
    function xGamma_Function(x: extended): extended;
  public
    { Public 錾 }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

const
  LONG_MAX = 2147483647;
  max_long_double_arg: extended = 1755.5;
  exp_g_o_sqrt_2pi = 6.23316569877722552586386e+3;
  e =  2.71828182845904523536028747;
  g =  9.65657815377331589457187;
  a: array[0..8] of extended = (
                                +1.14400529453851095667309e+4,
                                -3.23988020152318335053598e+4,
                                +3.50514523505571666566083e+4,
                                -1.81641309541260702610647e+4,
                                +4.63232990536666818409138e+3,
                                -5.36976777703356780555748e+2,
                                +2.28754473395181007645155e+1,
                                -2.17925748738865115560082e-1,
                                +1.08314836272589368860689e-4
                               );

var
  LDBL_EPSILON : extended;

//==============================================================================
// ֐̌vZ
// ֐̌vZ̐́AK}֐̃y[WQƂĂB
//==============================================================================

function TForm1.Duplication_Formula(two_x: extended): extended;
var
  x: extended;
  g: extended;
  n: integer;
begin
  x := 0.5 * two_x;
  n := round(two_x) - 1;
  g := power(2.0, two_x - 1.0 - n);
  g := g * power(2, n);
  g := g / sqrt(pi);
  g := g * xGamma_Function(x);
  g := g * xGamma_Function(x + 0.5);
  result := g;
end;

function TForm1.xGamma(x: extended): extended;
var
  xx: extended;
  temp: extended;
  n, i: integer;
  xx2 : extended;
begin
  if x < 1 then xx := x + 1 else xx := x;
  n := sizeof(a) div sizeof(extended);
  if x > max_long_double_arg then begin
    result := maxextended;
    exit;
  end;
  if x > 900 then begin
    result := Duplication_Formula(x);
    exit;
  end;
  temp := 0;
  for i := n - 1 downto 0 do temp := temp + a[i] / (xx + i);
  temp := temp + 1;
  xx2 := (xx - 0.5) / 2;
  temp := temp * (power((g + xx - 0.5) / e, xx2) / exp_g_o_sqrt_2pi );
  temp := temp * power((g + xx - 0.5) / e, xx2);      // X64 I[o[t[΍
  if x < 1 then result := temp / x
           else result := temp;
end;

function TForm1.xGamma_Function(x: extended): extended;
var
  sin_x: extended;
  rg: extended;
  ix: int64;
begin
  if x > 0 then
    if x <= max_long_double_arg then begin
      result := xGamma(x);
      exit;
    end
    else begin
      result := maxextended;
      exit;
    end;
  if x > -LONG_MAX then begin
    ix := round(x);
    if x = ix then begin
      result := maxextended;
      exit;
    end;
  end;
  sin_x := sin(pi * x);
  if sin_x = 0 then begin
    result := maxextended;
    exit;
  end;
  if x < - max_long_double_arg - 1 then begin
    result := 0;
    exit;
  end;
  rg := xGamma(1 - x) * sin_x / pi;
  if rg <> 0 then result := 1 / rg
             else result := maxextended;
end;

//==============================================================================
// function Dirichlet_Eta_Functionis: double): double;
//
// ̃[`́AEta֐vZ܂
// eta(s)=i1-2^(1-s))zeta(s)
// ŁAzeta()̓[}[[^֐łB
// s >= 0̏ꍇ
//   eta(s)= Sum(-1)^(k-1) (1 / k^s)
//   ŁAvk = 1,...ŉZ@eta(1)= lni2j
// s < 0̏ꍇ
//   ˎgp邱ƂɒӂĂB
// s << 0̏ꍇA
//   etaisǰX́A傫傫AۓȂ܂B
//   ͑傫ȐΌ덷쐬\܂B
//==============================================================================
function TForm1.Dirichlet_Eta_Function(s: double): double;
var
  x : double;
begin
  x := xDirichlet_Eta_Function(s);
  if abs(x) < MAXDouble then result := x
  else
    if x > 0 then result := Maxdouble
             else result := -Maxdouble;
end;

function TForm1.xDirichlet_Eta_Function(s: extended): extended;
begin
  if s >= 64.0 then begin
    result := 1;
    exit;
  end;
  result :=  1.0 + xDirichlet_Eta_Star_Function(s);
end;

function TForm1.Dirichlet_Eta_Star_Function(s: double): double;
var
  x : double;
begin
  x := xDirichlet_Eta_Star_Function(s);
  if abs(x) < MAXDouble then result := x
  else
    if  x > 0.0 then result := MAXdouble
                else result := -MAXdouble;
end;

function TForm1.xDirichlet_Eta_Star_Function(s : extended): extended;
var
  reflection_coefficients : extended;
begin
  if s >= 0.0 then
    if s < 18.0 then begin
      result := Alternating_Series_Convergence_Acceleration(s);
      exit;
    end
    else begin
      result := Sum_Reverse_Order(s);
      exit;
    end;

    // Use the reflection formula  s < 0

  reflection_coefficients := Reflection_Coefficient(s);
  result := reflection_coefficients * xDirichlet_Eta_Star_Function(1.0 - s)
          + (reflection_coefficients - 1.0);
end;


//==============================================================================
// function Reflection_Coefficient(s: extended): extended;
//
// fBÑC[^֐̔ˌ͎̂ƂłB
//   eta(s)= {[2*(1-2^(1-s))/(1-2^s)]*(1-sj
//                      /(2*pi)^(1-s)*cos((1-s*pi/2)}*eta(1-s)
//     ŁAs < 0 ͕̎łB
//
// ̃[`͌W:
//    [2*(1-2^(1-s))/(1-2^s)]*(1-s)/(2*pi)^(1-s)*cos((1-s)*pi/2) Ԃ܂B
//@@ŁAs < 0͕̎łB
// 
//  s : extended
//    ŁAs < 0łB
//
// ߂lF
// @@etaisj= c * etai1-sj ƂȂW c
//==============================================================================
function TForm1.Reflection_Coefficient(s: extended): extended;
var
  one_s : extended;
  k : integer;
  v : extended;
  c : extended;
  x, temp: extended;
begin
  one_s := 1.0 - s;
  k := trunc(one_s / 4.0);
  v := one_s - 4.0 * k;
  c := cos(v * pi / 2);
  if abs(c) < 1.8 * LDBL_EPSILON then begin
    result := 0.0;
    exit;
  end;
  temp := power(2.0, one_s);
  x := (1.0 - temp) / (temp - 2.0);
  x := x + x;
  x := x * c;
  x := x * xGamma_Function(one_s);
  x := x / power(pi, one_s);
  result := x;
end;

//==============================================================================
// function Alternating_Series_Convergence_Acceleration(s: extended): extended;
//
// F
//  ̃[`́As <= 18 Eta֐vZ܂B
//  Ǝ܂B
//
// eNjbNF
//   ㋉^ꂽꍇFv Sum(-1)^ka[k]AK = 0 ...ōv܂B
//@@ŁAׂĂkɑ΂a [k]> 0łB̊֐݂ꍇ w(x)
//    [0,1]ŁAa[k] = I[0,1] x^kw(x)dx
//    x^kw(x)xɊւ 0  1 ܂Őϕ
//    sum(-1)^ka[k] = sum(-1)^kI[0,1] x^kw(x) dx
//                  = I[0,1] Sum(-1)^kx^kw(x) dx
//                  = I[0,1] w(x)/(1 + x) dx
//
// S = I [0,1] w(x)/i1 + xjdx Ƃ܂B
// P(-1)I= 0ƂȂ悤Ȏ n ̑ P(x)^ꂽꍇ
//      P(x)= p[0] - p[1] x + p[2] x^2 - ...i-1^np [n] x^n
//         S[n] = I[0,1] (P(-1) - P(x)) w(x) / (1+xjdx / P(-1) `܂B
// ɁA  S[n] = S-I [0,1] Pixjwixj/i1 + xjdx / Pi-1j
// ܂ |S[n] -S| <= I[0,1] |P(x)| w(x) / (1 + x) dx / |P(x)|
//        |S[n] -S| <= (M / |P(-1)|) SA M = max |P(x)|Ax in [0,1]
//
// 㐔̌
//          S[n] = (1/ P(-1)) Sum Sum p[j][(-1)^k a[k]]A
// ŏ̍v k = 0A...An-1A2Ԗڂ̍v j = k + 1A...An
//
// d[k] = Sum p[j] sumed for j = k + 1A...An for k = n-1A...A0A
// S[n] = (1/c[0]jSum d[k] (-1)^ka [k] summed k = 0A...An-1
//
// ŃvOꂽ P(x) ̑I (-1)^n * T*[n](x) łBA
// T*[n] 1 n  Vtgꂽ`FrVFtłB
// P(x) ̂悤ɑIƁA
//     P(x) = Sum {(n / (n + j)) C(n+jA2j) 4^j (-x)^j}A
// v j = 0A...An and C(n+j,2j) = (n+j)!/[(2j)!(nj)!]
// and M = 1AP(-1) >= (1/2) (3 + sqrt(8))^nB
//
// F
// s : extended
//   s < 18
//
// ߂lF
//  eta(s)
//==============================================================================
function TForm1.Alternating_Series_Convergence_Acceleration(s: extended): extended;
const
  d: array[0..28] of extended = (
     1.362725501650887306817e+21, 1.362725501650887306816e+21,
     1.362725501650887305248e+21, 1.362725501650886896000e+21,
     1.362725501650844334208e+21, 1.362725501648488235008e+21,
     1.362725501568066715648e+21, 1.362725499718371770368e+21,
     1.362725469310199922688e+21, 1.362725096810094788608e+21,
     1.362721590926752350208e+21, 1.362695647390018306048e+21,
     1.362542007743905005568e+21, 1.361803869444099801088e+21,
     1.358896740140251611136e+21, 1.349437033675348770816e+21,
     1.323863206542645919744e+21, 1.266218975223368122368e+21,
     1.157712186857668739072e+21, 9.872015194258554224640e+20,
     7.640581139674368573440e+20, 5.220333434317674905600e+20,
     3.061506212814840135680e+20, 1.496014168469232680960e+20,
     5.884825485587356057600e+19, 1.781624012587768217600e+19,
     3.882102878793367552000e+18, 5.404319552844595200000e+17,
     3.602879701896396800000e+16
    );
var
  term: array[0..27] of extended;
  sum: extended;
  k: integer;
begin
  k := 1;
  while  k <= 28 do begin
    term[k-1] := d[k] * power(k + 1, -s);
    inc(K);
    term[k-1] := -d[k] * power(k + 1,-s);
    inc(k);
  end;
  sum := term[27];
  for k := 26 downto 0 do sum := sum + term[k];
  sum := sum / d[0];
  result := -sum;
end;

//==============================================================================
// function Sum_Reverse_Orderis: extended): extended;
//
// ̃[`́A
//     s >= 18  Eta֐vZ܂(power()ւ̌Ăяo28)Kvł)
//
//  eta(s) = Sum (-1^(k-1) (1/ k^s
//     v k = 1A...ōv܂B
//
//@n񂪌݂ɂȂĂ邽߁AΌ덷͍ŏɖꂽ̐ΒlłB
//@@
//@v̂́AA鉺ƏvZ邱ƂɂČ肳܂
//  2̋EȂ܂Łi؂̂ăG[ŁjvZ
//@ɁAŏIʂŌ̍獇vAŏ̍ŏI܂B
// i͏młjB
//
// F
//  s: extended;
//
// ߂lF
//  eta(s)
//==============================================================================
function TForm1.Sum_Reverse_Order(s: extended): extended;
var
  term: array[0..29] of extended;
  lower_bound : extended;
  upper_bound : extended;
  sum : extended;
  k : integer;
begin
  lower_bound := -power(2.0, -s);
  term[0] := lower_bound;
  term[1] := power(3.0, -s);
  upper_bound := term[1] + lower_bound;
  if lower_bound = upper_bound then begin
    result := upper_bound;
    exit;
  end;
  k := 4;
  while k < 32 do begin
    term[k - 2] := -power(k, -s);
    lower_bound := upper_bound + term[k - 2];
    if lower_bound = upper_bound then break;
    inc (k);
    term[k - 2] := power(k, -s);
    upper_bound := lower_bound + term[ k- 2];
    if lower_bound = upper_bound then break;
    inc(k);
  end;
  k := k - 2;
  sum := term[k];
  for K := k - 1 downto 0 do sum := sum +  term[k];
  result := sum;
end;

//==============================================================================
// fBÑC[^֐vZƃOt\
//==============================================================================
procedure TForm1.BitBtn1Click(Sender: TObject);
const
  n = 200;
var
  ans, anss, s: double;
  min, max : double;
  x, dx : double;
  ch : integer;
begin
  val(LabeledEdit1.Text, s, ch);
  if ch <> 0 then begin
    application.MessageBox('','S̒lɊԈႢ܂B',0);
    exit;
  end;
  Memo1.Clear;
  ans := Dirichlet_Eta_Function(s);
  Memo1.Lines.Append('s = ' + floatTostr(s));
  Memo1.Lines.Append(' (s)   = ' + floatTostr(ans));
  anss := Dirichlet_Eta_Star_Function(s);
  Memo1.Lines.Append('*(s) = (s) - 1');;
  Memo1.Lines.text := Memo1.Lines.text + ' *(s) = ' + floatTostr(anss);
  Series1.Clear;
  Series2.Clear;
  min := s - 10;
  if min < - 1 then begin
    min := s - 0.5;
    if s >= 0 then
      if s - min < 10 then min := -0.5;
  end;
  max := s + 10;
  dx := (max - min) / n;
  for ch := 0 to n do begin
    x := dx * ch + min;
    anss := Dirichlet_Eta_Function(x);
    Series1.AddXY(x, anss);
  end;
  Series2.AddXY(s, ans);
end;

//==============================================================================
// LDBL_EPSILON̐ݒ
//==============================================================================
procedure TForm1.FormCreate(Sender: TObject);
var
  one, tmp : extended;
begin
  one := 1;
  tmp := 1;
  repeat
    tmp := tmp / 2;
  until one + tmp = 1;
  LDBL_EPSILON := tmp * 2;
  memo1.Clear;
end;

end.
